import os
import time
import shutil
from typing import Sequence, List, Union
from tempfile import TemporaryDirectory

from pydantic import Field, DirectoryPath
from mkite_core.models import JobInfo, JobResults, Status
from mkite_engines.settings import EngineSettings

from .base import BaseEngine, BaseProducer, BaseConsumer


LOCAL_QUEUE_PREFIX = "queue-"


class LocalEngineSettings(EngineSettings):
    root_path: DirectoryPath = Field(
        os.path.expanduser("~/queue"),
        description="Where the jobs to be built will be placed",
    )
    move: bool = Field(
        False,
        description="If True, moves the paths when pushing",
    )
    delay: float = Field(
        2.0,
        description="Delay between considering the local folder as ready",
    )


class LocalEngine(BaseEngine):
    """Engine to implement a queue system using filesystem folders"""

    SETTINGS_CLS = LocalEngineSettings

    def __init__(
        self,
        root_path: os.PathLike,
        move: bool = False,
        return_abspath: bool = True,
        queue_prefix: str = LOCAL_QUEUE_PREFIX,
        delay: float = 2.0,
    ):
        self.root_path = os.path.abspath(root_path)
        self.mkdir(self.root_path)
        self.move = move
        self.return_abspath = return_abspath
        self.qprefix = queue_prefix
        self.delay = delay

    def __len__(self):
        return len(self.queues)

    def __repr__(self):
        n = len(self.queues)
        plural = "s" if n > 1 else ""
        return f"<{self.__class__.__name__} @ {self.root_path} ({n} queue{plural})>"

    def add_queue(self, name: str):
        if isinstance(name, Status):
            name = name.value

        if not self.is_queue(name):
            name = self.format_queue_name(name)

        if name not in self.queues:
            self.setup_path(name)

    def abspath(self, folder: os.PathLike):
        return os.path.join(self.root_path, folder)

    def setup_path(self, name: str):
        path = self.abspath(name)
        self.mkdir(path)
        return path

    def mkdir(self, path: os.PathLike):
        if os.path.exists(path):
            return

        os.mkdir(path)

    def delete(self, path: os.PathLike):
        if not os.path.isabs(path):
            path = self.abspath(path)

        elif not str(path).startswith(self.root_path):
            raise ValueError(
                "Not allowed to delete a file outside \
                of the path of the queue."
            )

        if not os.path.exists(path):
            raise FileNotFoundError(f"Path {path} does not exist")

        if not os.path.isdir(path):
            os.remove(path)
            return

        shutil.rmtree(path)

    def is_path(self, path: str):
        return os.path.exists(path)

    def item_path(self, queue: str, item: os.PathLike):
        name = os.path.basename(item)
        queue = self.format_queue_name(queue)
        dst = self.abspath(queue)
        return os.path.join(dst, name)

    def copy_path(self, queue: str, item: os.PathLike):
        dst = self.item_path(queue, item)

        if os.path.isdir(item):
            shutil.copytree(item, dst)
        else:
            shutil.copy(item, dst)

        return dst

    def move_path(self, queue: str, item: os.PathLike):
        dst = self.item_path(queue, item)
        shutil.move(item, dst)

        return dst

    def list_path(self, path: os.PathLike):
        return [entry for entry in os.listdir(path) if not entry.startswith(".")]

    def remove_path(self, item: os.PathLike):
        shutil.rmtree(item)

    def get_queue_path(self, queue: str) -> str:
        if isinstance(queue, Status):
            queue = queue.value

        if self.remove_queue_prefix(queue) not in self.queues:
            raise ValueError(f"Invalid queue {queue}")

        queue = self.format_queue_name(queue)
        return self.abspath(queue)

    def list_queue(self, queue: str) -> List[str]:
        """Get `n` items from the queue"""
        path = self.get_queue_path(queue)
        return self.list_path(path)

    def list_queue_names(self) -> List[str]:
        return [
            self.remove_queue_prefix(f)
            for f in self.list_path(self.root_path)
            if os.path.isdir(os.path.join(self.root_path, f)) and self.is_queue(f)
        ]

    @property
    def queues(self) -> List[str]:
        return self.list_queue_names()

    def set_status(self, key: str, status: str = Status.DOING.value):
        src = self.abspath(key)
        return self.move_path(status, src)


class LocalProducer(LocalEngine, BaseProducer):
    """Producer that submits a folder to a new directory"""

    def push(
        self,
        queue: str,
        item: os.PathLike,
        add_queue: bool = True,
    ):
        if queue not in self.queues and add_queue:
            self.add_queue(queue)

        if not self.is_path(item):
            raise ValueError(f"Cannot submit {item}: invalid type")

        if self.move:
            dst = self.move_path(queue, item)
        else:
            dst = self.copy_path(queue, item)

        return dst

    def push_info(
        self,
        queue: str,
        info: Union[JobInfo, JobResults],
        status=Status.READY.value,
        name: str = None,
    ):
        if name is None and self.is_info(info):
            name = info.uuid

        if not name.endswith(".json"):
            name = name + ".json"

        with TemporaryDirectory() as tmp:
            path = os.path.join(tmp, name)
            info.to_json(path)
            dst = self.push(queue, path)

        return dst


class LocalConsumer(LocalEngine, BaseConsumer):
    """Consumer that uses folders as queue"""

    def is_valid(self, path: os.PathLike):
        if not os.path.exists(path):
            return False

        if os.path.basename(path).startswith("."):
            return False

        mtime = os.path.getmtime(path)
        now = time.time()

        return (now - mtime) > self.delay

    def get(self, queue: str) -> (str, str):
        """Get an item from the queue"""
        path = self.get_queue_path(queue)

        entries = self.list_path(path)

        if len(entries) == 0:
            return None, None

        item = entries[0]
        key = os.path.join(queue, item)

        if self.return_abspath:
            return key, os.path.join(path, item)

        return key, item

    def get_n(self, queue: str, n: int = 1000) -> (str, str):
        """Get `n` items from the queue"""
        path = self.get_queue_path(queue)

        i = 0
        for entry in os.scandir(path):
            if not self.is_valid(entry.path):
                continue

            if i >= n:
                break

            key = entry.name

            if self.return_abspath:
                item = os.path.join(path, key)
            else:
                item = key

            yield key, item

            i = i + 1

        return None, None

    def get_info(
        self, queue: str, info_cls=JobInfo
    ) -> (str, Union[JobInfo, JobResults]):
        """Get a JobInfo from the queue"""
        key, path = self.get(queue)
        if path is None:
            return None, None

        info = info_cls.from_json(path)

        return path, info



import click

from mkite_engines.cli.redis import redis


class MkiteEnginesGroup(click.Group):
    pass


@click.command(cls=MkiteEnginesGroup)
def kiteng():
    """Command line interface for mkite_core"""


kiteng.add_command(redis)

if __name__ == "__main__":
    kiteng()



import json
import redis

from typing import List, Union, Optional
from pydantic import ConfigDict, Field, DirectoryPath, BaseModel
from mkite_engines.settings import EngineSettings
from mkite_core.models import JobInfo, JobResults, Status

from .base import BaseEngine, BaseProducer, BaseConsumer


class RedisEngineSettings(EngineSettings):
    host: str = Field(
        "localhost",
        description="hostname or IP address of the Redis server",
    )
    port: int = Field(
        6379,
        description="port of the Redis server",
    )
    username: Optional[str] = Field(
        "abc", 
        description="username of the Redis server"
    )
    password: str = Field(
        "abc",
        description="password of the Redis server",
    )
    ssl: bool = Field(False, description="whether to use ssl when connecting")
    ssl_cert_reqs: str = Field(
        "required",
        description="requirements of ssl certificates",
    )
    model_config = ConfigDict(env_prefix="REDIS_", case_sensitive=False)


class RedisInfoSchema(BaseModel):
    msg: str
    status: str

    def items(self):
        return self.model_dump().items()


class RedisEngine(BaseEngine):
    SETTINGS_CLS = RedisEngineSettings

    def __init__(
        self,
        host: str,
        port: int,
        password: str = "abc",
        queue_prefix: str = "queue:",
        **kwargs,
    ):
        self.redis_kwargs = {
            "host": host,
            "port": port,
            "password": password,
            "health_check_interval": 30,
            **kwargs,
        }
        self.qprefix = queue_prefix
        self._r = None

    @property
    def r(self):
        if self._r is None:
            self._r = self._get_new_redis()

        try:
            self._r.ping()

        except redis.ConnectionError:
            del self._r
            self._r = self._get_new_redis()

        return self._r

    def _get_new_redis(self):
        return redis.Redis(**self.redis_kwargs)

    def list_queue(self, queue: str) -> List[str]:
        queue = self.format_queue_name(queue)
        items = self.r.lrange(queue, 0, -1)
        return [i.decode() for i in items]

    def list_queue_names(self) -> List[str]:
        queues = self.r.keys(self.qprefix + "*")
        queues = [k.decode() for k in queues]
        queues = [self.remove_queue_prefix(k) for k in queues]
        return queues

    def add_queue(self, name: str):
        """Empty queues do not have to be created in Redis.
        This method exists for compatibility with other engines.
        """
        pass

    def set_status(self, key: str, status: str = Status.DOING.value):
        self.r.hset(key, "status", status)

    def delete(self, key: str):
        self.r.delete(key)


class RedisProducer(RedisEngine, BaseProducer):
    def push(self, queue: str, item: str, left: bool = True):
        queue = self.format_queue_name(queue)
        if left:
            return self.r.lpush(queue, item)
        return self.r.rpush(queue, item)

    def push_info(
        self,
        queue: str,
        info: Union[JobInfo, JobResults],
        status=Status.READY.value,
    ):
        msg = info.encode()
        key = str(info.uuid)

        data = RedisInfoSchema(
            msg=msg,
            status=status,
        )

        self.r.hset(key, mapping=data)
        return self.push(queue, key)


class RedisConsumer(RedisEngine, BaseConsumer):
    def get(self, queue: str) -> (str, str):
        if queue not in self.list_queue_names():
            return None, None

        queue = self.format_queue_name(queue)
        key = self.r.lpop(queue).decode()
        msg = self.r.hget(key, "msg")

        return key, msg



from .base import BaseEngine, BaseProducer, BaseConsumer, EngineRoles
from .local import LocalEngine, LocalProducer, LocalConsumer
from .redis import RedisEngine, RedisProducer, RedisConsumer
from .instantiate import get_engine_class, instantiate_from_dict, instantiate_from_path

PUBLISHERS = {
    "local": LocalProducer,
    "redis": RedisProducer,
}

CONSUMERS = {
    "local": LocalConsumer,
    "redis": RedisConsumer,
}

ENGINES = {
    "local": LocalEngine,
    "redis": RedisEngine,
}



import os
from enum import Enum
from typing import List, Dict, Union
from abc import ABC, abstractmethod
from mkite_core.external import load_config

from mkite_core.models import JobInfo, JobResults
from .settings import EngineSettings


class EngineError(Exception):
    pass


class EngineRoles(Enum):
    producer = "producer"
    consumer = "consumer"


class BaseEngine(ABC):
    """Manages the flow of information to/from mkite/mkwind and their processes.
    As mkite and mkwind are not coupled directly, an intermediate engine has to
    act as the middleware for their communication.
    """

    SETTINGS_CLS = EngineSettings

    def __init__(
        self,
        *args,
        queue_prefix: str = "queue:",
        **kwargs
    ):
        self.qprefix = queue_prefix

    @classmethod
    def from_settings(cls, settings: EngineSettings) -> "BaseEngine":
        """Creates the engine from a pydantic EngineSettings"""
        return cls(**settings.model_dump())

    @classmethod
    def from_env(cls) -> "BaseEngine":
        """Creates the engine from environmental variables"""
        settings = cls.SETTINGS_CLS()
        return cls.from_settings(settings)

    @classmethod
    def from_file(cls, filename: os.PathLike) -> "BaseEngine":
        """Creates the engine from a configuration file"""
        settings = cls.SETTINGS_CLS.from_file(filename)
        return cls.from_settings(settings)

    def format_queue_name(self, queue: str):
        if self.is_queue(queue):
            return queue

        return self.qprefix + queue

    def remove_queue_prefix(self, queue: str):
        return queue.replace(self.qprefix, "")

    def is_queue(self, key: str) -> bool:
        return key.startswith(self.qprefix)

    @abstractmethod
    def list_queue(self, queue: str) -> List[str]:
        """Fully lists a given queue"""

    def list_all_queues(self) -> List[str]:
        """Fully lists all queues"""
        items = []
        for queue in self.list_queue_names():
            items += self.list_queue(queue)

        return items

    @abstractmethod
    def list_queue_names(self) -> List[str]:
        """Lists all names of all queues in the engine"""

    def is_info(self, item) -> bool:
        """Returns True if `item` is an instance of JobInfo or JobResults"""
        return isinstance(item, (JobInfo, JobResults))

    @abstractmethod
    def add_queue(self, name: str):
        """Creates a new queue in the engine"""

    @abstractmethod
    def delete(self, key: str):
        """Deletes an item indexed by `key` from the engine"""


class BaseProducer(BaseEngine):
    @abstractmethod
    def push(self, queue: str, item: str):
        """Adds the given item to the queue"""

    def push_info(self, queue: str, info: JobInfo):
        return self.push(queue, info.encode())


class BaseConsumer(BaseEngine):
    @abstractmethod
    def get(self, queue: str) -> (str, str):
        """Get an item from the queue"""

    def pop(self, queue: str) -> (str, str):
        key, item = self.get(queue)

        if key is not None:
            self.delete(key)

        return key, item

    def get_n(self, queue: str, n: int = 1000) -> (str, str):
        """Get n items from the queue"""
        i = 0
        while i < n:
            key, item = self.get(queue)
            if not item:
                break

            yield key, item
            i = i + 1

        return None, None

    def get_info(self, queue: str, info_cls=JobInfo) -> (str, Union[JobInfo, JobResults]):
        """Get a JobInfo from the queue"""
        key, item = self.get(queue)

        if item is None:
            return None, None

        return key, info_cls.decode(item)



import os
import unittest as ut
from unittest.mock import patch
from pkg_resources import resource_filename

from mkite_core.models import Status
from mkite_engines.base import EngineError, EngineRoles
from mkite_engines.settings import EngineSettings
from mkite_engines.local import LocalProducer, LocalConsumer

from mkite_engines.instantiate import (
    get_engine_class,
    instantiate_from_dict,
    instantiate_from_path,
)

SETTINGS_PATH = resource_filename("mkite_engines.tests.configs", "local.yaml")


class TestInstantiate(ut.TestCase):
    def setUp(self):
        self.module = "mkite_engines.local"

    def get_local_dict(self):
        return {
            "_module": self.module,
            "root_path": ".",
        }

    def test_get_class(self):
        cls = get_engine_class(self.module, EngineRoles.producer)
        self.assertEqual(cls, LocalProducer)

        cls = get_engine_class(self.module, EngineRoles.consumer)
        self.assertEqual(cls, LocalConsumer)

    def test_from_dict(self):
        settings = self.get_local_dict()
        obj = instantiate_from_dict(settings, EngineRoles.producer)
        self.assertIsInstance(obj, LocalProducer)

    def test_from_path(self):
        obj = instantiate_from_path(SETTINGS_PATH, EngineRoles.producer)
        self.assertIsInstance(obj, LocalProducer)



import os
from mkite_core.external import load_config
from pydantic import Field, DirectoryPath, FilePath
from pydantic_settings import BaseSettings, SettingsConfigDict


class EngineSettings(BaseSettings):
    """Wraps and obtains all settings for the environmental variables"""
    model_config = SettingsConfigDict(case_sensitive=False)

    @classmethod
    def from_file(cls, filename: FilePath):
        data = load_config(filename)

        if "_module" in data:
            data.pop("_module")

        return cls(**data)



import os
import uuid
import fakeredis
import unittest as ut
from unittest.mock import patch

from mkite_core.models import JobInfo, JobResults, Status
from mkite_engines.redis import (
    RedisEngineSettings,
    RedisInfoSchema,
    RedisEngine,
    RedisProducer,
    RedisConsumer,
)


def get_info():
    return JobInfo(
        job={"uuid": str(uuid.uuid4())},
        recipe={"name": "test"},
        options={"param1": 1},
        inputs=[],
    )


def get_fake_redis(**kwargs):
    return fakeredis.FakeStrictRedis(**kwargs)


class TestRedisEngine(ut.TestCase):
    def setUp(self):
        self.settings = RedisEngineSettings()
        self.engine = RedisEngine.from_settings(self.settings)
        self.engine._r = get_fake_redis(**self.engine.redis_kwargs)

    def tearDown(self):
        self.engine.r.flushall()

    def test_format_queue_name(self):
        queue = "test"
        expected = self.engine.qprefix + queue
        returned = self.engine.format_queue_name(queue)
        self.assertEqual(expected, returned)

        queue = f"{self.engine.qprefix}test2"
        expected = queue
        returned = self.engine.format_queue_name(queue)
        self.assertEqual(expected, returned)

    def test_is_queue(self):
        queue = "test"
        self.assertFalse(self.engine.is_queue(queue))

        queue = f"{self.engine.qprefix}test2"
        self.assertTrue(self.engine.is_queue(queue))

    def test_list_queue(self):
        queue = f"{self.engine.qprefix}test"

        expected = []
        returned = self.engine.list_queue(queue)
        self.assertEqual(returned, expected)

        expected = ["a", "b", "c"]
        self.engine.r.rpush(queue, *expected)
        returned = self.engine.list_queue(queue)
        self.assertEqual(returned, expected)

    def test_list_queue_names(self):
        returned = self.engine.list_queue_names()
        self.assertEqual(returned, [])

        N_QUEUES = 4
        queues = [f"{self.engine.qprefix}{i}" for i in range(N_QUEUES)]
        for q in queues:
            self.engine.r.rpush(q, "item")

        expected = [f"{i}" for i in range(N_QUEUES)]
        returned = self.engine.list_queue_names()
        self.assertEqual(returned, expected)


class TestRedisProducer(ut.TestCase):
    def setUp(self):
        self.settings = RedisEngineSettings()
        self.prod = RedisProducer.from_settings(self.settings)
        self.prod._r = get_fake_redis(**self.prod.redis_kwargs)

    def tearDown(self):
        self.prod.r.flushall()

    def test_push(self):
        queue = f"{self.prod.qprefix}test"
        item = "test_item"

        self.prod.push(queue, item)
        returned = self.prod.r.lpop(queue).decode()

        self.assertEqual(returned, item)

    def test_push_info(self):
        info = get_info()
        key = info.uuid

        queue = f"{self.prod.qprefix}test"

        self.prod.push_info(queue, info)

        expected = [key]
        returned = [item.decode() for item in self.prod.r.lrange(queue, 0, -1)]
        self.assertEqual(expected, returned)

        status = self.prod.r.hget(key, "status").decode()
        self.assertEqual(status, Status.READY.value)

        msg = self.prod.r.hget(key, "msg")
        self.assertEqual(msg, info.encode())


class TestRedisConsumer(ut.TestCase):
    def setUp(self):
        self.settings = RedisEngineSettings()
        self.prod = RedisProducer.from_settings(self.settings)

        # connects both fakeredis caches
        self.prod._r = get_fake_redis(**self.prod.redis_kwargs)
        self.cons = RedisConsumer.from_settings(self.settings)
        self.cons._r = self.prod._r

    def tearDown(self):
        self.prod.r.flushall()
        self.cons.r.flushall()

    def test_get(self):
        queue = "test"

        info = get_info()
        key = info.uuid
        self.prod.push_info(queue, info)

        status = Status.READY.value
        returned_key, msg = self.cons.get(queue)

        self.assertEqual(key, returned_key)

        info_str = info.encode()
        self.assertEqual(msg, info_str)

        new_status = self.cons.r.hget(key, "status").decode()
        self.assertEqual(status, new_status)

    def test_get_info(self):
        queue = "test"

        info = get_info()
        key = info.uuid
        self.prod.push_info(queue, info)

        new_key, new_info = self.cons.get_info(queue)
        self.assertEqual(key, new_key)
        self.assertEqual(info, new_info)



import click

from mkite_engines.redis import RedisEngine


def summary(engine):
    queues = engine.list_queue_names()
    for q in queues:
        res = engine.list_queue(q)
        print(f"{q}: {len(res)}")


@click.command("redis")
@click.option(
    "-s",
    "--settings",
    type=str,
    default=None,
    help="path to the yaml file configuring the Redis engine",
)
@click.option(
    "-i",
    "--interactive",
    is_flag=True,
    default=False,
    help="If True, opens an ipython shell",
)
def redis(settings, interactive):
    from IPython import embed

    engine = RedisEngine.from_file(settings)
    summary(engine)

    if interactive:
        r = engine.r

        embed()



import os
from mkite_core.external import load_config

from .base import EngineError, EngineRoles


def get_engine_class(module: str, role: EngineRoles):
    """Obtains the engine class given by `module` and `role`.
    For example, obtaining the class for a LocalProducer
    would requiring passing the following arguments:

        module="mkite_engines.local", role="producer"

    Arguments:
        module (str): namespace of the engine module
        role (str): whether the role of the engine is producer or consumer.

    Returns:
        engine: class of type EngineRole
    """
    engine = module.split(".")[-1]
    clsname = engine.capitalize() + role.value.capitalize()

    _module = __import__(module, globals(), locals(), [clsname], 0)
    if not hasattr(_module, clsname):
        raise EngineError(f"{clsname} does not exist in module {module}")

    _cls = getattr(_module, clsname)
    return _cls


def instantiate_from_dict(settings: dict, role: EngineRoles, **kwargs):
    _module_key: str = "_module"
    _settings = {
        **settings,
        **kwargs,
    }
    if _module_key not in settings:
        raise EngineError(
            "Engine settings do not specify which engine to use."
            + f"Please specify a `{_module_key}` for the engine."
        )

    module = _settings.pop(_module_key)

    cls = get_engine_class(module, role)

    return cls(**_settings)


def instantiate_from_path(path: os.PathLike, role: EngineRoles, **kwargs):
    data = load_config(path)
    return instantiate_from_dict(data, role, **kwargs)
